"""The standardized, inter-changable way to represent specific data in Himena."""

from himena.standards._base import BaseMetadata, read_metadata, write_metadata

__all__ = ["BaseMetadata", "read_metadata", "write_metadata"]
